"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.startServer = void 0;
const theme_check_common_1 = require("@shopify/theme-check-common");
const vscode_languageserver_1 = require("vscode-languageserver");
const ClientCapabilities_1 = require("../ClientCapabilities");
const codeActions_1 = require("../codeActions");
const commands_1 = require("../commands");
const completions_1 = require("../completions");
const diagnostics_1 = require("../diagnostics");
const DocumentHighlightsProvider_1 = require("../documentHighlights/DocumentHighlightsProvider");
const documentLinks_1 = require("../documentLinks");
const documents_1 = require("../documents");
const formatting_1 = require("../formatting");
const hover_1 = require("../hover");
const JSONLanguageService_1 = require("../json/JSONLanguageService");
const LinkedEditingRangesProvider_1 = require("../linkedEditingRanges/LinkedEditingRangesProvider");
const RenameProvider_1 = require("../rename/RenameProvider");
const RenameHandler_1 = require("../renamed/RenameHandler");
const utils_1 = require("../utils");
const uri_1 = require("../utils/uri");
const version_1 = require("../version");
const CachedFileSystem_1 = require("./CachedFileSystem");
const Configuration_1 = require("./Configuration");
const safe_1 = require("./safe");
const defaultLogger = () => { };
/**
 * The `git:` VFS does not support the `fs.readDirectory` call and makes most things break.
 * `git` URIs are the ones you'd encounter when doing a git diff in VS Code. They're not
 * real files, they're just a way to represent changes in a git repository. As such, I don't
 * think we want to sync those in our document manager or try to offer document links, etc.
 *
 * A middleware would be nice but it'd be a bit of a pain to implement.
 */
const hasUnsupportedDocument = (params) => {
    return ('textDocument' in params &&
        'uri' in params.textDocument &&
        typeof params.textDocument.uri === 'string' &&
        params.textDocument.uri.startsWith('git:'));
};
/**
 * This code runs in node and the browser, it can't talk to the file system
 * or make requests. Stuff like that should be injected.
 *
 * In browser, theme-check-js wants these things:
 *   - fileExists(path)
 *   - defaultTranslations
 *
 * Which means we gotta provide 'em from here too!
 */
function startServer(connection, { fs: injectedFs, loadConfig, log = defaultLogger, jsonValidationSet, themeDocset: remoteThemeDocset, fetchMetafieldDefinitionsForURI, }) {
    const fs = new CachedFileSystem_1.CachedFileSystem(injectedFs);
    const fileExists = (0, theme_check_common_1.makeFileExists)(fs);
    const clientCapabilities = new ClientCapabilities_1.ClientCapabilities();
    const configuration = new Configuration_1.Configuration(connection, clientCapabilities);
    const documentManager = new documents_1.DocumentManager(fs, connection, clientCapabilities, getModeForURI, isValidSchema);
    const diagnosticsManager = new diagnostics_1.DiagnosticsManager(connection);
    const documentLinksProvider = new documentLinks_1.DocumentLinksProvider(documentManager, findThemeRootURI);
    const codeActionsProvider = new codeActions_1.CodeActionsProvider(documentManager, diagnosticsManager);
    const onTypeFormattingProvider = new formatting_1.OnTypeFormattingProvider(documentManager, async function setCursorPosition(textDocument, position) {
        if (!clientCapabilities.hasShowDocumentSupport)
            return;
        connection.sendRequest(vscode_languageserver_1.ShowDocumentRequest.type, {
            uri: textDocument.uri,
            takeFocus: true,
            selection: {
                start: position,
                end: position,
            },
        });
    });
    const linkedEditingRangesProvider = new LinkedEditingRangesProvider_1.LinkedEditingRangesProvider(documentManager);
    const documentHighlightProvider = new DocumentHighlightsProvider_1.DocumentHighlightsProvider(documentManager);
    const renameProvider = new RenameProvider_1.RenameProvider(documentManager);
    const renameHandler = new RenameHandler_1.RenameHandler(connection, clientCapabilities, documentManager, findThemeRootURI);
    async function findThemeRootURI(uri) {
        const rootUri = await (0, theme_check_common_1.findRoot)(uri, fileExists);
        const config = await loadConfig(rootUri, fs);
        return config.rootUri;
    }
    const getMetafieldDefinitionsForRootUri = (0, theme_check_common_1.memoize)((0, theme_check_common_1.makeGetMetafieldDefinitions)(fs), (rootUri) => rootUri);
    const getMetafieldDefinitions = async (uri) => {
        const rootUri = await findThemeRootURI(uri);
        return getMetafieldDefinitionsForRootUri(rootUri);
    };
    // These are augmented here so that the caching is maintained over different runs.
    const themeDocset = new theme_check_common_1.AugmentedThemeDocset(remoteThemeDocset);
    const runChecks = (0, utils_1.debounce)((0, diagnostics_1.makeRunChecks)(documentManager, diagnosticsManager, {
        fs,
        loadConfig,
        themeDocset,
        jsonValidationSet,
        getMetafieldDefinitions,
    }), 100);
    const getTranslationsForURI = async (uri) => {
        const rootURI = await findThemeRootURI(uri);
        const theme = documentManager.theme(rootURI);
        const getDefaultTranslations = (0, theme_check_common_1.makeGetDefaultTranslations)(fs, theme, rootURI);
        const [defaultTranslations, shopifyTranslations] = await Promise.all([
            getDefaultTranslations(),
            themeDocset.systemTranslations(),
        ]);
        return { ...shopifyTranslations, ...defaultTranslations };
    };
    const getSchemaTranslationsForURI = async (uri) => {
        const rootURI = await findThemeRootURI(uri);
        const theme = documentManager.theme(rootURI);
        const getDefaultSchemaTranslations = (0, theme_check_common_1.makeGetDefaultSchemaTranslations)(fs, theme, rootURI);
        return getDefaultSchemaTranslations();
    };
    const getSnippetDefinitionForURI = async (uri, snippetName) => {
        const rootUri = await findThemeRootURI(uri);
        const snippetURI = theme_check_common_1.path.join(rootUri, 'snippets', `${snippetName}.liquid`);
        const snippet = documentManager.get(snippetURI);
        if (!snippet || snippet.type !== theme_check_common_1.SourceCodeType.LiquidHtml || (0, theme_check_common_1.isError)(snippet.ast)) {
            return undefined;
        }
        return snippet.getLiquidDoc(snippetName);
    };
    const snippetFilter = ([uri]) => /\.liquid$/.test(uri) && /snippets/.test(uri);
    const getSnippetNamesForURI = (0, safe_1.safe)(async (uri) => {
        const rootUri = await findThemeRootURI(uri);
        const snippetUris = await (0, theme_check_common_1.recursiveReadDirectory)(fs, rootUri, snippetFilter);
        return snippetUris.map(uri_1.snippetName);
    }, []);
    const getThemeSettingsSchemaForURI = (0, safe_1.safe)(async (uri) => {
        const rootUri = await findThemeRootURI(uri);
        const settingsSchemaUri = theme_check_common_1.path.join(rootUri, 'config', 'settings_schema.json');
        const contents = await fs.readFile(settingsSchemaUri);
        const json = (0, theme_check_common_1.parseJSON)(contents);
        if ((0, theme_check_common_1.isError)(json) || !Array.isArray(json)) {
            throw new Error('Settings JSON file not in correct format');
        }
        return json;
    }, []);
    async function getModeForURI(uri) {
        const rootUri = await (0, theme_check_common_1.findRoot)(uri, fileExists);
        const config = await loadConfig(rootUri, fs);
        return config.context;
    }
    const getThemeBlockNames = (0, safe_1.safe)(async (uri, includePrivate) => {
        const rootUri = await findThemeRootURI(uri);
        const blocks = await fs.readDirectory(theme_check_common_1.path.join(rootUri, 'blocks'));
        const blockNames = blocks.map(([uri]) => theme_check_common_1.path.basename(uri, '.liquid'));
        if (includePrivate) {
            return blockNames;
        }
        return blockNames.filter((blockName) => !blockName.startsWith('_'));
    }, []);
    async function getThemeBlockSchema(uri, name) {
        const rootUri = await findThemeRootURI(uri);
        const blockUri = theme_check_common_1.path.join(rootUri, 'blocks', `${name}.liquid`);
        const doc = documentManager.get(blockUri);
        if (!doc || doc.type !== theme_check_common_1.SourceCodeType.LiquidHtml) {
            return;
        }
        return doc.getSchema();
    }
    // Defined as a function to solve a circular dependency (doc manager & json
    // lang service both need each other)
    async function isValidSchema(uri, jsonString) {
        return jsonLanguageService.isValidSchema(uri, jsonString);
    }
    const jsonLanguageService = new JSONLanguageService_1.JSONLanguageService(documentManager, jsonValidationSet, getSchemaTranslationsForURI, getModeForURI, getThemeBlockNames, getThemeBlockSchema);
    const completionsProvider = new completions_1.CompletionsProvider({
        documentManager,
        themeDocset,
        getTranslationsForURI,
        getSnippetNamesForURI,
        getThemeSettingsSchemaForURI,
        log,
        getThemeBlockNames,
        getMetafieldDefinitions,
    });
    const hoverProvider = new hover_1.HoverProvider(documentManager, themeDocset, getMetafieldDefinitions, getTranslationsForURI, getThemeSettingsSchemaForURI, getSnippetDefinitionForURI);
    const executeCommandProvider = new commands_1.ExecuteCommandProvider(documentManager, diagnosticsManager, clientCapabilities, runChecks, connection);
    const fetchMetafieldDefinitionsForWorkspaceFolders = async (folders) => {
        if (!fetchMetafieldDefinitionsForURI)
            return;
        for (let folder of folders) {
            const mode = await getModeForURI(folder.uri);
            if (mode === 'theme') {
                fetchMetafieldDefinitionsForURI(folder.uri);
            }
        }
    };
    connection.onInitialize((params) => {
        clientCapabilities.setup(params.capabilities, params.initializationOptions);
        jsonLanguageService.setup(params.capabilities);
        configuration.setup();
        const fileOperationRegistrationOptions = {
            filters: [
                {
                    pattern: {
                        glob: '**/*.{liquid,json}',
                    },
                },
                {
                    pattern: {
                        glob: '**/assets/*',
                    },
                },
            ],
        };
        const result = {
            capabilities: {
                textDocumentSync: {
                    change: vscode_languageserver_1.TextDocumentSyncKind.Full,
                    save: true,
                    openClose: true,
                },
                codeActionProvider: {
                    codeActionKinds: [...codeActions_1.CodeActionKinds],
                },
                completionProvider: {
                    triggerCharacters: ['.', '{{ ', '{% ', '<', '/', '[', '"', "'", ':'],
                },
                documentOnTypeFormattingProvider: {
                    firstTriggerCharacter: ' ',
                    moreTriggerCharacter: ['{', '%', '-', '>'],
                },
                documentLinkProvider: {
                    resolveProvider: false,
                    workDoneProgress: false,
                },
                documentHighlightProvider: true,
                linkedEditingRangeProvider: true,
                renameProvider: {
                    prepareProvider: true,
                },
                executeCommandProvider: {
                    commands: [...commands_1.Commands],
                },
                hoverProvider: {
                    workDoneProgress: false,
                },
                workspace: {
                    workspaceFolders: {
                        supported: true,
                        changeNotifications: true,
                    },
                    fileOperations: {
                        didRename: fileOperationRegistrationOptions,
                        didCreate: fileOperationRegistrationOptions,
                        didDelete: fileOperationRegistrationOptions,
                    },
                },
            },
            serverInfo: {
                name: 'theme-language-server',
                version: version_1.VERSION,
            },
        };
        return result;
    });
    connection.onInitialized(() => {
        log(`[SERVER] Let's roll!`);
        configuration.fetchConfiguration();
        configuration.registerDidChangeCapability();
        configuration.registerDidChangeWatchedFilesNotification({
            watchers: [
                {
                    globPattern: '**/.shopify/*',
                },
            ],
        });
        if (clientCapabilities.hasWorkspaceFoldersSupport) {
            connection.workspace.getWorkspaceFolders().then(async (folders) => {
                if (!folders)
                    return;
                fetchMetafieldDefinitionsForWorkspaceFolders(folders);
            });
            connection.workspace.onDidChangeWorkspaceFolders(async (params) => {
                fetchMetafieldDefinitionsForWorkspaceFolders(params.added);
            });
        }
    });
    connection.onDidChangeConfiguration((_params) => {
        configuration.clearCache();
    });
    connection.onDidOpenTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri, text, version } = params.textDocument;
        documentManager.open(uri, text, version);
        if (await configuration.shouldCheckOnOpen()) {
            runChecks([uri]);
        }
        // The objective at the time of writing this is to make {Asset,Snippet}Rename
        // fast when you eventually need it.
        //
        // I'm choosing the textDocument/didOpen notification as a hook because
        // I'm not sure we have a better solution than this. Yes we have the
        // initialize request with the workspace folders, but you might have opened
        // an app folder. The root of a theme app extension would probably be
        // at ${workspaceRoot}/extensions/${appExtensionName}. It'd be hard to
        // figure out from the initialize request params.
        //
        // If we open a file that we know is liquid, then we can kind of guarantee
        // we'll find a theme root and we'll preload that.
        if (await configuration.shouldPreloadOnBoot()) {
            const rootUri = await findThemeRootURI(uri);
            documentManager.preload(rootUri);
        }
    });
    connection.onDidChangeTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri, version } = params.textDocument;
        documentManager.change(uri, params.contentChanges[0].text, version);
        if (await configuration.shouldCheckOnChange()) {
            runChecks([uri]);
        }
        else {
            // The diagnostics may be stale! Clear em!
            diagnosticsManager.clear(params.textDocument.uri);
        }
    });
    connection.onDidSaveTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri } = params.textDocument;
        fs.readFile.invalidate(uri);
        fs.stat.invalidate(uri);
        if (await configuration.shouldCheckOnSave()) {
            runChecks([uri]);
        }
    });
    connection.onDidCloseTextDocument((params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri } = params.textDocument;
        documentManager.close(uri);
        diagnosticsManager.clear(uri);
    });
    connection.onDocumentLinks(async (params) => {
        if (hasUnsupportedDocument(params))
            return [];
        return documentLinksProvider.documentLinks(params.textDocument.uri);
    });
    connection.onCodeAction(async (params) => {
        return codeActionsProvider.codeActions(params);
    });
    connection.onExecuteCommand(async (params) => {
        await executeCommandProvider.execute(params);
    });
    connection.onCompletion(async (params) => {
        var _a;
        if (hasUnsupportedDocument(params))
            return [];
        return ((_a = (await jsonLanguageService.completions(params))) !== null && _a !== void 0 ? _a : (await completionsProvider.completions(params)));
    });
    connection.onHover(async (params) => {
        var _a;
        if (hasUnsupportedDocument(params))
            return null;
        return (_a = (await jsonLanguageService.hover(params))) !== null && _a !== void 0 ? _a : (await hoverProvider.hover(params));
    });
    connection.onDocumentOnTypeFormatting(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return onTypeFormattingProvider.onTypeFormatting(params);
    });
    connection.onDocumentHighlight(async (params) => {
        if (hasUnsupportedDocument(params))
            return [];
        return documentHighlightProvider.documentHighlights(params);
    });
    connection.onPrepareRename(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return renameProvider.prepare(params);
    });
    connection.onRenameRequest(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return renameProvider.rename(params);
    });
    connection.languages.onLinkedEditingRange(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return linkedEditingRangesProvider.linkedEditingRanges(params);
    });
    connection.onDidChangeWatchedFiles(async (params) => {
        if (params.changes.length === 0)
            return;
        for (const change of params.changes) {
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(change.uri));
            fs.readFile.invalidate(change.uri);
            fs.stat.invalidate(change.uri);
            if (change.uri.endsWith('metafields.json')) {
                const rootUri = await findThemeRootURI(change.uri);
                getMetafieldDefinitionsForRootUri.invalidate(rootUri);
            }
        }
    });
    connection.workspace.onDidCreateFiles((params) => {
        const triggerUris = params.files.map((fileCreate) => fileCreate.uri);
        for (const { uri } of params.files) {
            // When a file is created, to make sure preload isn't invalidated, we add
            // an empty string to the document manager for the new URI.
            if (!documentManager.has(uri))
                documentManager.open(uri, '', undefined);
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(uri)); // Since there's a new file there
            fs.readFile.invalidate(uri); // Since this is no longer an error
            fs.stat.invalidate(uri); // Since this is no longer an error
        }
        // MissingAssets/MissingSnippet should be rerun when a new file exists
        // since the file creation might invalidate the error.
        runChecks.force(triggerUris);
    });
    connection.workspace.onDidRenameFiles(async (params) => {
        const triggerUris = params.files.map((fileRename) => fileRename.newUri);
        // Behold the cache invalidation monster
        for (const { oldUri, newUri } of params.files) {
            // When a file is renamed, we paste the content of the old file into the
            // new file in the document manager. We don't need to invalidate preload
            // because that's the only thing that changed.
            documentManager.rename(oldUri, newUri);
            // When a file is renamed, readDirectory to the parent folder is invalidated.
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(oldUri));
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(newUri));
            // When a file is renamed, readFile and stat for both the old and new URIs are invalidated.
            fs.readFile.invalidate(oldUri);
            fs.readFile.invalidate(newUri);
            fs.stat.invalidate(oldUri);
            fs.stat.invalidate(newUri);
        }
        // We should complete refactors before running theme check
        await renameHandler.onDidRenameFiles(params);
        // MissingAssets/MissingSnippet should be rerun when a file is renamed
        // since the file rename might invalidate an error.
        runChecks.force(triggerUris);
    });
    connection.workspace.onDidDeleteFiles((params) => {
        const triggerUris = params.files.map((fileDelete) => fileDelete.uri);
        for (const { uri } of params.files) {
            // When a file is deleted, we remove it from the document manager.
            // Don't need to invalidate preload because that operation covers it.
            documentManager.delete(uri);
            // Since the file is gone, we invalidate the parent directory's readDirectory.
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(uri));
            // Since the file is gone, we invalidate the readFile and stat for the URI.
            fs.readFile.invalidate(uri);
            fs.stat.invalidate(uri);
        }
        // MissingAssets/MissingSnippet should be rerun when a file is deleted
        // since the file rename might cause an error.
        runChecks.force(triggerUris);
    });
    connection.listen();
}
exports.startServer = startServer;
//# sourceMappingURL=startServer.js.map